/*******************************************************************************
 *                            C++ Verimag
 *    
 *
 *------------------------------------------------------------------------------
 * 	            Distributed and Complex Systems - Gires
 *------------------------------------------------------------------------------
 * clock.h
 *------------------------------------------------------------------------------
 *  Real-time clocks management for BIP2 Engine implementation
 *  + class Clock       : clock management
 *------------------------------------------------------------------------------
 * Created by Jacques COMBAZ on Fri Nov 21 2008
 *
 ******************************************************************************/

#ifndef CLOCK_H
#define CLOCK_H

#include "bip-time.h"

/**************************************
 * Class: GenericClock
 * Description:
 *
 *
 ****************************************/

class GenericClock {

  /*** Protected data members ***/
  protected:
    char*        mName;              /* clock name */

  /*** Constructors & Dectructors ***/
  public:
    GenericClock();
    GenericClock(const char*);
    virtual ~GenericClock();

  /*** Protected Data-Access Methods ***/
  public:
    inline void     setName(char* name) { mName = name; }

  /*** Public Data-Access Methods ***/
  public:
    inline char*    name()        const { return mName; }


  /*** Public Utility Methods ***/
  public:
    virtual BIP::Time  time()                     const = 0; /* return the current value of the clock */
            BIP::Time  drift(const GenericClock*) const;     /* return the difference between two clocks */
            BIP::Time  drift(const GenericClock&) const;     /* return the difference between two clocks */
};


/**************************************
 * Class: PlatformClock
 * Description:
 *
 *
 ****************************************/

class PlatformClock: public GenericClock {

  /*** Constructors & Dectructors ***/
  public:
    PlatformClock(const char*);
    virtual ~PlatformClock();

  /*** Public Utility Methods ***/
  public:
    virtual BIP::Time time() const      = 0;
    virtual void wait(const BIP::Time&) = 0;
};


/**************************************
 * Class: SimuClock
 * Description:
 *
 *
 ****************************************/

class SimuClock: public PlatformClock {

  /*** Protected data members ***/
  protected:
    BIP::Time mTime;

  /*** Constructors & Dectructors ***/
  public:
    SimuClock(const char*);
    virtual ~SimuClock();

  /*** Protected Data-Access Methods ***/
  protected:
    inline  void setTime(const BIP::Time time) { mTime = time; }

  /*** Public Utility Methods ***/
  public:
            BIP::Time time() const;
            void wait(const BIP::Time&);
};


/**************************************
 * Class: LogicalClock
 * Description:
 *
 *
 ****************************************/

class LogicalClock: public GenericClock {

  /*** Protected data members ***/
  protected:
    PlatformClock& mPlatformClock; /* reference to the platform clock     */
    BIP::Time      mTime;          /* store the logical value of the time */

  /*** Constructors & Dectructors ***/
  public:
    LogicalClock(const char*, PlatformClock&);
    virtual ~LogicalClock();

  /*** Public Data-Access Methods ***/
  public:
    inline PlatformClock& platformClock() const { return mPlatformClock; }


  /*** Protected Data-Access Methods ***/
  protected:
    inline void           setTime(const BIP::Time& time) { mTime = time; }


  /*** Public Utility Methods ***/
  public:
    inline void           reset() { setTime(BIP::Time(0)); }; /* for exploration only, should not be used elsewhere */
           BIP::Time      time() const;
           void           wait(const BIP::Time&);
           void           update();
};



/**************************************
 * Class: Clock
 * Description:
 *
 *
 ****************************************/

class Clock: public GenericClock {

  /*** Protected data members ***/
  protected:
    const LogicalClock& mLogicalClock;      /* global time                                      */
          BIP::Time     mResetTime;         /* time of the last reset w.r.t. mGlobalClock       */
          bool          mStopped;           /* true iff the clock is stopped (stop progressing) */
          BIP::Time     mIntermediateTime;  /* used to store intermediate time when stopped     */

  /*** Constructors & Dectructors ***/
  public:
    Clock(const char*, const LogicalClock&);
    Clock(const char*, const LogicalClock*); // for compatibilty with old generated code, to be removed
    virtual ~Clock();

  /*** Protected Data-Access Methods ***/
  protected:
    inline const LogicalClock& logicalClock()                         const { return mLogicalClock; }
    inline       BIP::Time     resetTime()                            const { return mResetTime; }
    inline       void          setResetTime(const BIP::Time& time)          { mResetTime = time; }
    inline       BIP::Time     intermediateTime()                     const { return mIntermediateTime; }
    inline       void          setIntermediateTime(const BIP::Time& time)   { mIntermediateTime = time; }
    inline       void          setStopped(bool stopped)                     { mStopped = stopped; }

  /*** Public Data-Access Methods ***/
  public:
    inline bool          isStopped()                    const { return mStopped; }
    inline void          setTime(const BIP::Time& time)       { mResetTime = -time; }; /* for exploration only, assume that the current */
                                                                                       /* logical clock time is zero */

  /*** Public Utility Methods ***/
  public:
           BIP::Time     time() const;
           void          reset();                      /* reset the clock  */
           void          stop();                       /* stop progressing */
           void          go();                         /* stop pausing     */
};

#endif

