/*******************************************************************************
 *                            C++ Verimag
 *    
 *
 *------------------------------------------------------------------------------
 * 	            Distributed and Complex Systems - Gires
 *------------------------------------------------------------------------------
 * constraint.h
 *------------------------------------------------------------------------------
 *  Time constraints implementation for BIP2 Engine
 *  + class Interval    : time interval with urgency type
 *  + class Constraint  : list for time intervals
 *------------------------------------------------------------------------------
 * Created by Jacques COMBAZ on Fri Nov 21 2008
 *
 ******************************************************************************/

#ifndef CONSTRAINT_H
#define CONSTRAINT_H

#include "bip-time.h"

#include <vector>
#include <iostream>

class GenericClock;

/* defines the types of urgency used in BIP, ordered by priority level */
enum Urgency { LAZY = 0, DELAYABLE = 1, EAGER = 2 };

/**************************************
 * Class: Interval
 * Description:
 *
 *
 ****************************************/
class Interval {
  /*** Protected data members ***/
  protected:
    BIP::Time    mLow;         /* lower bound of the interval */
    BIP::Time    mHigh;        /* upper bound of the interval */
    bool         mDominated;   /* dominated in a priority rule */
    Urgency      mUrgency;     /* urgency type */
    bool         mObservable;  /* observability w.r.t. the environment */
    

  /*** Constructors & Dectructors ***/
  public:
    Interval(const BIP::Time = BIP::Time::MIN, const BIP::Time = BIP::Time::MAX, Urgency = LAZY, bool = false, bool = true);
    Interval(const Interval &);
    Interval(const BIP::Time&, const BIP::Time&, const Interval &);
    Interval(const BIP::Time&, const BIP::Time&, const Interval &, bool);
    Interval(const Interval &, bool, Urgency = LAZY, bool = true);
    virtual ~Interval();


  /*** Public Data-Access Methods ***/
  public:
    inline BIP::Time    low()                                    const { return mLow; }
    inline BIP::Time    high()                                   const { return mHigh; }
    inline void         setLow(const BIP::Time& low)                   { mLow  = low; }
    inline void         setHigh(const BIP::Time& high)                 { mHigh = high; }
           void         applyDelay(const BIP::Time& delay);
    inline bool         dominated()                              const { return mDominated; }
    inline void         setDominated(bool dominated)                   { mDominated = dominated; }
    inline Urgency      urgency()                                const { return mUrgency; }
    inline void         setUrgency(Urgency urgency)                    { mUrgency = urgency; }
    inline void         setMaxUrgency(Urgency urgency)                 { if (urgency > mUrgency) mUrgency = urgency; }
    inline bool         observable()                             const { return mObservable; }
    inline void         setObservable(bool observable)                 { mObservable  = observable; }
    inline void         setMaxObservable(bool observable)              { mObservable |= observable; }
 

  /*** Public Utility Methods ***/
  public:
    inline bool      in(const BIP::Time& time)             const { return (low() <= time) && (time <= high()); }
    inline bool      inside(const BIP::Time& time)         const { return (low() <  time) && (time <  high()); }
    inline bool      in(const Interval& interval)          const { return (low() <= interval.low()) && (interval.high() <= high()); }
    inline bool      inside(const Interval& interval)      const { return (low() <  interval.low()) && (interval.high() <  high()); }
    inline void      setEmpty()                                  { setLow(BIP::Time::MAX); setHigh(BIP::Time::MIN); }
    inline bool      empty()                               const { return high() < low(); }
    inline bool      singleton()                           const { return low() == high(); }
 
           Interval& operator=(const Interval&);
           Interval& operator&=(const Interval&);
           Interval& operator-=(const Interval&);

    inline       Interval& operator+=(const BIP::Time& delay)          { setLow(low() + delay); setHigh(high() + delay); return *this; }
    inline       Interval& operator-=(const BIP::Time& delay)          { setLow(low() - delay); setHigh(high() - delay); return *this; }
    inline const Interval  operator+ (const BIP::Time& delay)          { Interval interval = *this; interval += delay; return interval; }
    inline const Interval  operator- (const BIP::Time& delay)          { Interval interval = *this; interval -= delay; return interval; }

    inline bool      operator< (const Interval& interval) const { return (interval.low() <  low()) && (high() <  interval.high()); }
    inline bool      operator<=(const Interval& interval) const { return (interval.low() <= low()) && (high() <= interval.high()); }
    inline bool      operator> (const Interval& interval) const { return (low() <  interval.low()) && (interval.high() <  high()); }
    inline bool      operator>=(const Interval& interval) const { return (low() <= interval.low()) && (interval.high() <= high()); }
    inline bool      operator==(const Interval& interval) const { return (low() == interval.low()) && (interval.high() == high()); }
    inline bool      operator!=(const Interval& interval) const { return (low() != interval.low()) || (interval.high() != high()); }

    friend std::ostream&  operator<<(std::ostream &, const Interval&);
};


/**************************************
 * Class: Constraint
 * Description:
 *
 *
 ****************************************/
class Constraint {

  /*** Protected data members ***/
  protected:
    GenericClock*         mClock;        /* constraint w.r.t. a given clock */
    GenericClock*         mDisplayClock; /* clock used for << operator (implicit conversion) */
    bool                  mDomination;   /* if true, intervals are set dominated when applying -= */
    BIP::Time             mDelay;        /* delay for the -= operator */
    std::vector<Interval>  mIntervals;    /* ordered list of intervals */

  /*** Constructors & Dectructors ***/
  public:
    Constraint(const GenericClock* = NULL, Interval = Interval());
    Constraint(const Interval);
    Constraint(const Constraint&);
    Constraint(const GenericClock*, BIP::Time, BIP::Time, Urgency);
    virtual ~Constraint();


  /*** Protected Utility Methods ***/
  protected:
           int   before(const BIP::Time&) const;
           int   after(const BIP::Time&) const;
           int   position(const BIP::Time&, int&) const;
           int   position(const BIP::Time&) const;
           bool  intersect(const Interval&, int&, int&) const;
    inline void  erase(int index) { mIntervals.erase(mIntervals.begin() + index); }
           bool  erase(int, int);
    inline void  insert(int index, const Interval& interval) { if (!interval.empty()) mIntervals.insert(mIntervals.begin() + index, interval); }
           bool  setDominated(int, int, bool);
           void  setAllDominated(bool);
           bool  setUrgency(int, int, Urgency);
           bool  setMaxUrgency(int, int, Urgency);


  /*** Public Data-Access Methods ***/
  public:
    inline GenericClock* clock()                              const { return mClock; }
    inline void          setClock(const GenericClock* clock)        { mClock = (GenericClock*) clock; }
    inline GenericClock* displayClock()                       const { return mDisplayClock; }
    inline void          setDisplayClock(const GenericClock* clock) { mDisplayClock = (GenericClock*) clock; }
    inline bool          domination()                         const { return mDomination; }
    inline void          setDomination(bool domination)             { mDomination = domination; }
    inline BIP::Time     delay()                              const { return mDelay; }
    inline void          setDelay(const BIP::Time& delay)           { mDelay = delay; }

    inline bool      empty()                        const { return mIntervals.empty(); }
    inline int       size()                         const { return mIntervals.size(); }
           Interval  operator[](unsigned int index) const { return mIntervals[index]; }
           Interval& operator[](unsigned int index)       { return mIntervals[index]; }

  /*** Public Utility Methods ***/
  public:
    inline bool      in(const BIP::Time& time) const { return position(time) >= 0; }
           bool      in(const Interval &) const;
           bool      urgent(const BIP::Time&) const;                       /* return true iff the time constraint is urgent at time  */
           bool      existUrgency(int, int, Urgency) const;
           BIP::Time deadline(const BIP::Time&) const;                      /* return the next time urgencyPredicate() is true after time  */
           BIP::Time observableDeadline(const BIP::Time&) const;            /* return the next time urgencyPredicate(nextTime) && observable(nextTime) is true after time */
           BIP::Time next(const BIP::Time&) const;                          /* return the next time the guard is true after time           */
           bool      observable(const BIP::Time&) const;                    /* return true iff time is in an observable interval           */
           bool      check() const;                                         /* debug function */
           void      reset();                                               /* set the guard always true, and urgency LAZY */
    inline void      clear() { mIntervals.clear(); }                        /* remove all intervals (empty guard) */
           void      clean();                                               /* remove empty intervals, concatenate intervals when possible */

           void      applyDomination();                                     /* remove all dominated intervals */

           Constraint& operator-=(const Constraint&);    /* minus (set)  operator used for computing the priorities                  */ 
           Constraint& operator&=(const Constraint&);    /* intersection operator used for computing interactions between ports      */
           Constraint& operator|=(const Constraint&);    /* union        operator used for computing xPort (hierarchical connectors) */

           Constraint& operator+=(const BIP::Time&);     /* time translation operator */
           Constraint& operator-=(const BIP::Time&);     /* time translation operator */

    friend std::ostream&  operator<<(std::ostream &, const Constraint&);
    friend std::ostream&  operator<<(std::ostream &, const Constraint*);
};


#endif

