#ifndef CONNECTOR_H
#define CONNECTOR_H

#include "structure.h"
#include <stdarg.h>
#include <vector>
#include <unordered_map>
#include <cstdio>

class Connector;
class Compound;
class Port;
class Priority;

namespace std {
template <> 
struct hash<BitVector>
{
    size_t operator()(const BitVector &val) const;
};

inline size_t hash<BitVector>::operator()(const BitVector &val) const {
  size_t res(0);

//   std::cerr << "start: " << res << std::endl;
  for(BitVector::const_iterator it = val.begin(); it != val.end(); ++it) {
    res <<= 1;
    res += *it;
//     std::cerr << res << std::endl;
  }
//   std::cerr << "end: " << res << std::endl;

  return res;
}

}

/*********************************
 * Class: Interaction
 * Description: Interaction class
 *********************************/
class Interaction {
  /*** Protected data members ***/
  public:
    enum Status {UNKNOWN = 0, PENDING = 1, INHIBITED = 2, MAXIMAL = 3};

    typedef std::vector<Priority*> priorities_type;
    typedef std::vector<Interaction*> interactions_type;

  /*** Protected data members ***/
  protected:
    Connector *mConnector;  /* the parent Connector */
    BitVector mSynced;     /* the synced ports     */
    priorities_type mInhibitors; /* potential inhibitors */
    interactions_type mSuper;   /* super interactions   */
    Status                mStatus;

  /*** Constructors & Dectructors ***/
  public:
    Interaction(Connector*, const BitVector&);
    Interaction(const Interaction&);
    virtual ~Interaction();

  /*** Public Data-Access Methods ***/
  public:
    Connector* connector() const
    {
      return mConnector;
    }

    const BitVector& synced() const
    {
      return mSynced;
    }

    priorities_type& inhibitors()
    {
      return mInhibitors;
    }

    interactions_type& super()
    {
      return mSuper;
    }

    Status status() const               { return mStatus; }
    void setStatus(const Status s)      { mStatus = s; }

  /*** Public Utility Methods ***/
  public:
    void   setup();
    bool   equal(const Interaction*) const;
    bool   isSubset(const Interaction*, bool&) const;
    void   execute();
    Status evaluate();
    void   evaluateTC(Interaction*);
    static Interaction* Create(Connector*, const BitVector&);
    static Interaction* Create(Connector*, int size, ...);
    
    void print(FILE* file) const;
    void dump(FILE* file) const;

    //void   notifyDependentPorts();
    Interaction* copy() const;
 
    friend class Scheduler;
};

/*******************************
 * Class: Connector
 * Description: Connector class
 *******************************/
class Connector {
  /*** Public data members ***/
  public:
    enum Status {UNSET = 0, UNKNOWN = 1, ENABLED = 2, DISABLED = 3};

    typedef std::vector<Port*> ports_type;
    typedef std::vector<Connector*> connectors_type;
    typedef std::vector<Interaction*> interactions_type;
    typedef std::unordered_map<BitVector, Interaction> interactionsHashMap_type;

  /*** Private data members ***/
  private:
    static char LABEL[1024];

  /*** Protected data members ***/
  protected:
    char*              mName;          /* name of connector */
    Compound*          mHolder;        /* holder compound */
    ports_type mPorts;         /* port array */
    Port*              mxPort;         /* exported port */
    bool               mLegal;         /* legal i.e. complete or maximal */
    bool               mDominated;     /* dominated by priorities */
    Interaction*       mFeasibleInter; /* feasible interaction */
    bool               mSubconn;       /* used as a sub-connector
                                          in hierarcical connector */
    bool               mEval;          /* flag marking evaluated */
    connectors_type mLinks;    /* back pointer to higher links */
    Status             mStatus;        /* status wrt the scheduler */

    interactionsHashMap_type mStoredInteractions; /* store the interactions created through this connector */

  /*** Constructors & Dectructors ***/
  public:
    Connector(const char*, Compound*, const bool, const int, ...);
    Connector(const char*, Compound*, const bool); 
    virtual ~Connector();

  /*** Public Data-Access Methods ***/
  public:
    char* name() const
    {
      return mName;
    }

    Compound* holder() const
    {
      return mHolder;
    }

    ports_type& ports()
    {
      return mPorts;
    }

    const ports_type& ports() const
    {
      return mPorts;
    }

    Port*     xPort()     const { return mxPort; }
    void      setxPort(Port* p) { mxPort = p; }
    bool      legal()     const { return mLegal; }
    bool      dominated() const { return mDominated; }
    void      setDominated (bool dominated) { mDominated = dominated; }
    Interaction* feasibleInter() const { return mFeasibleInter; }
    void      setFeasibleInter(Interaction* i) { mFeasibleInter = i; }
    bool      isSubconn() const { return mSubconn; }
    bool      evaluated() const { return mEval; } 
    void      setEvaluated(bool b)   { mEval = b; }

    connectors_type& links()
    {
      return mLinks;
    }

    Status    status()    const { return mStatus; }
    void setStatus(const Status status)  { mStatus = status; }

   /*** Public Utility Methods ***/
   public:
     char*        label();    
     Interaction* evaluate(); 
     void         execute();
     void         createFeasibleInter();
     void         resetLink();
     void         resetDependentLinks();
     bool         synced()   const;
     virtual bool complete() const;
     virtual bool guard();
     bool         maximal()  const;
     virtual void up();
     virtual void dn();
     void         setVarPorts(const int, va_list);

     virtual void setup();
     virtual bool isBroadcast() const;
     virtual void setupBroadcast(const int);

     virtual bool allAtomsReady() const;
     virtual bool disabled() const;
     virtual bool unknown() const;

     bool         holderReady();
     void         holderGoBusy(int&);
     void         addLink(Connector*);

     virtual void dump(FILE* file) const;

     Interaction* getInteraction(const BitVector &syncedPorts);

   /*** Public Methods invoked from appln. ***/
   public:
     virtual void define() {}
};

#endif
